# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio CLI helpers module."""

from .cookiecutter_wrapper import CookiecutterWrapper
from .docker_helper import DockerHelper
from .env import env
from .filesystem import get_created_files



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

from ..helpers import env, filesystem
from ..helpers.process import run_cmd
from .local import LocalCommands
from .packages import PackagesCommands
from .steps import CommandStep, FunctionStep


class InstallCommands(LocalCommands):
    """Local installation commands."""

    def __init__(self, cli_config):
        """Constructor."""
        super(InstallCommands, self).__init__(cli_config)

    def install_py_dependencies(self, pre, dev=False):
        """Install Python dependencies."""
        # If not locked, lock. Then install.
        steps = []

        if PackagesCommands.is_locked().status_code > 0:
            steps.extend(PackagesCommands.lock(pre, dev))

        steps.extend(PackagesCommands.install_locked_dependencies(pre))

        return steps

    def update_instance_path(self):
        """Update path to instance in config."""
        # FIXME: Transform into steps.
        # Currently not possible because the second step (update instance
        # path) requires the ouptut of the previous step
        result = run_cmd(
            ['pipenv', 'run', 'invenio', 'shell', '--no-term-title',
                '-c', '"print(app.instance_path, end=\'\')"']
        )
        self.cli_config.update_instance_path(result.output.strip())

        result.output = "Instance path updated successfully."
        return result

    def symlink_project_file_or_folder(self, target):
        """Create symlink in instance pointing to project file or folder."""
        target_path = self.cli_config.get_project_dir() / target
        link_path = self.cli_config.get_instance_path() / target

        return filesystem.force_symlink(target_path, link_path)

    def install(self, pre, dev=False, flask_env='production'):
        """Development installation steps."""
        steps = self.install_py_dependencies(pre=pre, dev=dev)
        steps.append(
            FunctionStep(
                func=self.update_instance_path,
                message="Updating instance path..."
            )
        )
        steps.append(
            FunctionStep(
                func=self.symlink_project_file_or_folder,
                args={"target": 'invenio.cfg'},
                message=f"Symlinking 'invenio.cfg'..."
            )
            )
        steps.append(
            FunctionStep(
                func=self.symlink_project_file_or_folder,
                args={"target": 'templates'},
                message=f"Symlinking 'templates'..."
            )
            )
        steps.append(
            FunctionStep(
                func=self.symlink_project_file_or_folder,
                args={"target": 'app_data'},
                message=f"Symlinking 'app_data'..."
            )
        )
        steps.append(
            FunctionStep(
                func=self.update_statics_and_assets,
                args={"force": True, "flask_env": flask_env},
                message="Updating statics and assets..."
            )
        )

        return steps



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 Northwestern University.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module commands/local.py's tests."""

from os import environ
from pathlib import Path
from unittest.mock import Mock, call, patch

import pytest
from click import UsageError

from invenio_cli.commands import LocalCommands


@pytest.mark.skip()
@patch('invenio_cli.commands.local.run_cmd')
def test_install_py_dependencies(p_run_cmd, mock_cli_config):
    commands = LocalCommands(mock_cli_config)

    commands._install_py_dependencies(pre=True, lock=True)

    p_run_cmd.assert_any_call(
        ['pipenv', 'install', '--dev', '--pre']
    )

    commands._install_py_dependencies(pre=True, lock=False)

    p_run_cmd.assert_any_call(
        ['pipenv', 'install', '--dev', '--pre', '--skip-lock']
    )

    commands._install_py_dependencies(pre=False, lock=True)

    p_run_cmd.assert_any_call(['pipenv', 'install', '--dev'])

    commands._install_py_dependencies(pre=False, lock=False)

    p_run_cmd.assert_any_call(
        ['pipenv', 'install', '--dev', '--skip-lock'],
    )


@pytest.mark.skip()
@patch('invenio_cli.commands.local.PIPE')
@patch('invenio_cli.commands.local.run_cmd')
def test_update_instance_path(p_run_cmd, p_PIPE):
    cli_config = Mock()
    p_run_cmd.return_value = Mock(stdout='instance_dir')
    commands = LocalCommands(cli_config)

    commands._update_instance_path()

    p_run_cmd.assert_called_with(
        ['pipenv', 'run', 'invenio', 'shell', '--no-term-title',
            '-c', '"print(app.instance_path, end=\'\')"'],
        check=True, universal_newlines=True, stdout=p_PIPE
    )
    cli_config.update_instance_path.assert_called_with('instance_dir')


@pytest.mark.skip()
@patch('invenio_cli.helpers.filesystem.symlink')
def test_symlink_project_file_or_folder(p_symlink, mock_cli_config):
    commands = LocalCommands(mock_cli_config)
    file = 'invenio.cfg'

    commands._symlink_project_file_or_folder(file)

    p_symlink.assert_called_with(
        Path('project_dir/invenio.cfg'), Path('instance_dir/invenio.cfg'))

    folder = 'templates/'

    commands._symlink_project_file_or_folder(folder)

    p_symlink.assert_called_with(
        Path('project_dir/templates'), Path('instance_dir/templates'))


@pytest.mark.skip()
@patch('invenio_cli.helpers.filesystem.symlink')
def test_symlink_assets_templates(p_symlink, mock_cli_config):
    commands = LocalCommands(mock_cli_config)
    files_to_link = ['instance_dir/templates/template.js']

    commands._symlink_assets_templates(files_to_link)

    p_symlink.assert_called_with(
        Path('project_dir/templates/template.js'),
        Path('instance_dir/templates/template.js')
    )


@pytest.mark.skip()
@patch('invenio_cli.commands.local.copy_tree')
@patch('invenio_cli.commands.local.run_cmd')
def test_update_statics_and_assets(p_run_cmd, p_copy_tree, mock_cli_config):
    commands = LocalCommands(mock_cli_config)
    commands.update_statics_and_assets(force=True)

    expected_calls = [
        call(['pipenv', 'run', 'invenio', 'collect', '--verbose']),
        call(['pipenv', 'run', 'invenio', 'webpack', 'clean', 'create']),
        call(['pipenv', 'run', 'invenio', 'webpack', 'install']),
        call(['pipenv', 'run', 'invenio', 'webpack', 'build']),
    ]
    assert p_run_cmd.mock_calls == expected_calls
    p_copy_tree.assert_any_call(
        'project_dir/static', 'instance_dir/static'
    )
    p_copy_tree.assert_any_call(
        'project_dir/assets', 'instance_dir/assets'
    )

    # Reset for install=False assertions
    p_run_cmd.reset_mock()

    commands.update_statics_and_assets(force=False)

    expected_calls = [
        call(['pipenv', 'run', 'invenio', 'collect', '--verbose']),
        call(['pipenv', 'run', 'invenio', 'webpack', 'create']),
        call(['pipenv', 'run', 'invenio', 'webpack', 'build'])
    ]
    assert p_run_cmd.mock_calls == expected_calls


@pytest.mark.skip()
@patch('invenio_cli.commands.local.run_cmd')
@patch('invenio_cli.commands.local.DockerHelper')
def test_watch(p_docker_helper, p_run_cmd, mock_cli_config):
    LocalCommands(mock_cli_config).watch_assets()

    p_run_cmd.assert_called_with(
        ['pipenv', 'run', 'invenio', 'webpack', 'run', 'start']
    )


@pytest.mark.skip()
def test_install(mock_cli_config):
    commands = LocalCommands(mock_cli_config)
    commands._install_py_dependencies = Mock()
    commands._update_instance_path = Mock()
    commands._symlink_project_file_or_folder = Mock()
    commands.update_statics_and_assets = Mock()

    commands.install(False, False)

    commands._install_py_dependencies.assert_called_with(False, False)
    commands._update_instance_path.assert_called()
    expected_symlink_calls = [
        call('invenio.cfg'),
        call('templates'),
        call('app_data')
    ]
    assert (
        commands._symlink_project_file_or_folder.mock_calls ==
        expected_symlink_calls
    )
    commands.update_statics_and_assets.assert_called_with(
        force=True, flask_env='production')


@pytest.mark.skip()
@patch('invenio_cli.commands.local.DockerHelper')
@patch('invenio_cli.commands.local.run_cmd')
@patch('invenio_cli.helpers.process.popen')
def test_services(
        p_popen, p_run_cmd, p_docker_helper, mock_cli_config, mocked_pipe):
    commands = LocalCommands(mock_cli_config)

    p_popen.return_value = mocked_pipe
    commands.services(force=False)

    expected_setup_calls = [
        call(['pipenv', 'run', 'invenio', 'db', 'init', 'create']),
        call([
            'pipenv', 'run', 'invenio', 'files', 'location', 'create',
            '--default', 'default-location', 'instance_dir/data'
        ]),
        call([
            'pipenv', 'run', 'invenio', 'roles', 'create', 'admin'
        ]),
        call([
            'pipenv', 'run', 'invenio', 'access', 'allow',
            'superuser-access', 'role', 'admin'
        ]),
        call(['pipenv', 'run', 'invenio', 'index', 'init'])
    ]
    assert p_run_cmd.mock_calls == expected_setup_calls

    # Reset for install=False assertions
    p_run_cmd.reset_mock()

    commands.services(force=True)

    expected_force_calls = [
        call([
            'pipenv', 'run', 'invenio', 'shell', '--no-term-title', '-c',
            "import redis; redis.StrictRedis.from_url(app.config['CACHE_REDIS_URL']).flushall(); print('Cache cleared')"  # noqa
        ]),
        call([
            'pipenv', 'run', 'invenio', 'db', 'destroy', '--yes-i-know',
        ]),
        call([
            'pipenv', 'run', 'invenio', 'index', 'destroy', '--force',
            '--yes-i-know'
        ]),
        call([
            'pipenv', 'run', 'invenio', 'index', 'queue', 'init', 'purge',
        ])
    ]
    assert p_run_cmd.mock_calls == (
        expected_force_calls + expected_setup_calls
    )


@pytest.mark.skip()
@patch('invenio_cli.commands.local.run_cmd')
@patch('invenio_cli.commands.local.DockerHelper')
@patch('invenio_cli.helpers.process.popen')
def test_demo(
        p_popen, p_docker_helper, p_run_cmd, mock_cli_config, mocked_pipe):
    commands = LocalCommands(mock_cli_config)

    p_popen.return_value = mocked_pipe
    commands.demo()

    p_run_cmd.assert_called_with(
        ['pipenv', 'run', 'invenio', 'rdm-records', 'demo']
    )


@pytest.mark.skip()
@patch('invenio_cli.commands.local.DockerHelper')
@patch('invenio_cli.commands.local.run_cmd')
@patch('invenio_cli.commands.local.popen')
@patch('invenio_cli.helpers.process.popen')
def test_run(
        p_services_popen, p_commands_popen, p_run_cmd, p_docker_helper,
        mock_cli_config, mocked_pipe):
    commands = LocalCommands(mock_cli_config)
    p_services_popen.return_value = mocked_pipe

    host = '127.0.0.1'
    port = '5000'
    commands.run(host=host, port=port, debug=True)

    run_env = environ.copy()
    run_env['FLASK_ENV'] = 'development'
    run_env['INVENIO_SITE_HOSTNAME'] = f"{host}:{port}"
    expected_calls = [
        call([
            'pipenv', 'run', 'celery', '--app', 'invenio_app.celery', 'worker'
        ]),
        call([
            'pipenv', 'run', 'invenio', 'run', '--cert',
            'docker/nginx/test.crt', '--key', 'docker/nginx/test.key',
            '--host', '127.0.0.1', '--port', '5000'
        ], env=run_env),
        call().wait()
    ]
    assert p_commands_popen.mock_calls == expected_calls


@pytest.mark.skip()
@patch('pynpm.package.run_npm')
def test_link_js_module(p_run_npm, testpkg, mock_cli_config):

    LocalCommands(mock_cli_config).link_js_module(testpkg)

    expected_calls = [
        call(testpkg, 'run-script', npm_bin='npm', args=('link-dist',)),
        call('instance_dir/assets', 'link', npm_bin='npm', args=('testpkg',))
    ]

    assert expected_calls == p_run_npm.mock_calls


@pytest.mark.skip()
@patch('pynpm.package.run_npm')
def test_watch_js_module(p_run_npm, testpkg, mock_cli_config):

    LocalCommands(mock_cli_config).watch_js_module(
        testpkg, link=False)

    expected_calls = [
        call(testpkg, 'run-script', npm_bin='npm', args=('watch',)),
    ]
    assert expected_calls == p_run_npm.mock_calls


@pytest.mark.skip()
@patch('pynpm.package.run_npm')
def test_watch_js_module_w_build(p_run_npm, testpkg, mock_cli_config):

    LocalCommands(mock_cli_config).watch_js_module(testpkg, link=True)

    expected_calls = [
        call(testpkg, 'run-script', npm_bin='npm', args=('link-dist',)),
        call('instance_dir/assets', 'link', npm_bin='npm', args=('testpkg',)),
        call(testpkg, 'run-script', npm_bin='npm', args=('watch',)),
    ]
    assert expected_calls == p_run_npm.mock_calls


@pytest.mark.skip()
@patch('invenio_cli.commands.local.run_cmd')
def test_install_modules(p_run_cmd, mock_cli_config):
    commands = LocalCommands(mock_cli_config)

    """1 module"""
    modules = [Path().absolute()]
    commands.install_modules(modules)
    p_run_cmd.assert_called_with(
        ['pipenv', 'run', 'pip', 'install', '-e', modules[0]])

    """2 modules"""
    modules = [Path().absolute(), Path().absolute()]
    commands.install_modules(modules)
    p_run_cmd.assert_called_with(
        ['pipenv', 'run', 'pip', 'install', '-e', modules[0],
            '-e', modules[1]])

    """No modules"""
    modules = []
    with pytest.raises(UsageError):
        commands.install_modules(modules)
        cmd = ['pipenv', 'run', 'pip', 'install', '-e', modules]
        p_run_cmd(cmd)

    """Empty String"""
    """
    modules = [" "]
    with pytest.raises(click.UsageError):
        commands.install_modules(modules)
        cmd = ['invenio-cli', 'ext', 'module-install', modules[0]]
        print("comando")
        print(cmd)
        p_run_cmd(cmd)
    """

    """Invalid Path"""
    """
    modules = ["~/test/"]
    with pytest.raises(subprocess.CalledProcessError):
        commands.install_modules(modules)
        cmd = ['invenio-cli', 'ext', 'module-install', modules]
        p_run_cmd(cmd)
    """



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

import signal
from distutils.dir_util import copy_tree
from os import environ
from pathlib import Path
from subprocess import Popen as popen

import click

from ..helpers import env, filesystem
from ..helpers.process import ProcessResponse, run_interactive
from .commands import Commands
from .services import ServicesCommands


class LocalCommands(Commands):
    """Local CLI commands."""

    def __init__(self, cli_config):
        """Constructor."""
        super(LocalCommands, self).__init__(cli_config)

    def _symlink_assets_templates(self, files_to_link):
        """Symlink the assets folder."""
        assets = 'assets'
        click.secho('Symlinking {}...'.format(assets), fg='green')

        instance_path = self.cli_config.get_instance_path()
        project_dir = self.cli_config.get_project_dir()
        for file_path in files_to_link:
            file_path = Path(file_path)
            relative_path = file_path.relative_to(instance_path)
            target_path = project_dir / relative_path
            filesystem.force_symlink(target_path, file_path)

    def _copy_statics_and_assets(self):
        """Copy project's statics and assets into instance dir."""
        click.secho('Copying project statics and assets...', fg='green')

        static = 'static'
        src_dir = self.cli_config.get_project_dir() / static
        src_dir = str(src_dir)  # copy_tree below doesn't accept Path objects
        dst_dir = self.cli_config.get_instance_path() / static
        dst_dir = str(dst_dir)
        # using it for a different purpose then intended but very useful
        copy_tree(src_dir, dst_dir)

        assets = 'assets'
        src_dir = self.cli_config.get_project_dir() / assets
        src_dir = str(src_dir)
        dst_dir = self.cli_config.get_instance_path() / assets
        dst_dir = str(dst_dir)
        # The full path to the files that were copied is returned
        copied_files = copy_tree(src_dir, dst_dir)
        return copied_files

    def update_statics_and_assets(self, force, flask_env='production'):
        """High-level command to update less/js/images/... files.

        Needed here (parent) because is used by Assets and Install commands.
        """
        # Commands
        prefix = ['pipenv', 'run']
        collect_cmd = prefix + ['invenio', 'collect', '--verbose']
        clean_create_cmd = prefix + ['invenio', 'webpack', 'clean', 'create']
        create_cmd = prefix + ['invenio', 'webpack', 'create']
        install_cmd = prefix + ['invenio', 'webpack', 'install']
        build_cmd = prefix + ['invenio', 'webpack', 'build']

        with env(FLASK_ENV=flask_env):
            # Collect into statics/ and assets/ folder
            click.secho('Collecting statics and assets...', fg='green')
            run_interactive(collect_cmd, env={'PIPENV_VERBOSITY': "-1"})
            if force:
                run_interactive(
                    clean_create_cmd, env={'PIPENV_VERBOSITY': "-1"})

                # Installs in assets/node_modules/
                click.secho('Installing JS dependencies...', fg='green')
                run_interactive(install_cmd, env={'PIPENV_VERBOSITY': "-1"})
            else:
                run_interactive(create_cmd, env={'PIPENV_VERBOSITY': "-1"})

            # Symlink the instance's statics and assets
            copied_files = self._copy_statics_and_assets()
            self._symlink_assets_templates(copied_files)

            # Build project
            click.secho('Building assets...', fg='green')
            run_interactive(build_cmd, env={'PIPENV_VERBOSITY': "-1"})

        # FIXME: Refactor above to make use of proper error handling.
        return ProcessResponse(
                output="Assets and statics updated.",
                status_code=0,
        )

    def run(self, host, port, debug=True, services=True):
        """Run development server and celery queue."""

        def signal_handler(sig, frame):
            click.secho('Stopping server and worker...', fg='green')
            server.terminate()
            if services:
                worker.terminate()
            click.secho("Server and worker stopped...", fg="green")

        signal.signal(signal.SIGINT, signal_handler)

        if services:
            ServicesCommands(self.cli_config).ensure_containers_running()

            click.secho("Starting celery worker...", fg="green")
            worker = popen([
                'pipenv', 'run', 'celery', '--app',
                'invenio_app.celery', 'worker'
            ])

        click.secho("Starting up local (development) server...", fg='green')
        run_env = environ.copy()
        run_env['FLASK_ENV'] = 'development' if debug else 'production'
        run_env['INVENIO_SITE_HOSTNAME'] = f"{host}:{port}"
        server = popen([
            'pipenv', 'run', 'invenio', 'run', '--cert',
            'docker/nginx/test.crt', '--key', 'docker/nginx/test.key',
            '--host', host, '--port', port
        ], env=run_env)

        click.secho(
            f'Instance running!\nVisit https://{host}:{port}',
            fg='green')
        server.wait()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

import click

from ..helpers.cli_config import CLIConfig
from ..helpers.process import run_cmd

pass_cli_config = click.make_pass_decorator(CLIConfig, ensure=True)


def run_steps(steps, fail_message, success_message):
    """Run a series of steps."""
    for step in steps:
        click.secho(message=step.message, fg="green")
        result = step.execute()
        if result.status_code > 0:
            if result.error:
                fail_message = fail_message + f"\nErrors: {result.error}"
            if result.output:
                fail_message = fail_message + f"\nOutput: {result.output}"

            click.secho(fail_message, fg="red")
            exit(1)
        elif result.warning:
            if result.error:
                fail_message = fail_message + f"\nErrors: {result.error}"
            if result.output:
                fail_message = fail_message + f"\nOutput: {result.output}"

            click.secho(fail_message, fg="yellow")
        elif result.output:
            click.secho(message=result.output, fg="green")
    else:
        click.secho(message=success_message, fg="green")


def calculate_instance_path():
    """Caclulates instance path based on current venv."""
    result = run_cmd(
            ['pipenv', 'run', 'invenio', 'shell', '--no-term-title',
                '-c', '"print(app.instance_path, end=\'\')"']
        )
    return result.output.strip()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

import click

from ..commands import InstallCommands
from ..helpers.cli_config import CLIConfig
from .utils import pass_cli_config, run_steps


@click.command()
@click.option('--pre', default=False, is_flag=True,
              help='If specified, allows the installation of alpha releases')
@click.option('--dev/--no-dev', default=True, is_flag=True,
              help='Includes development dependencies.')
@click.option(
    '--production/--development', '-p/-d', default=False, is_flag=True,
    help='Production mode copies statics/assets. Development mode symlinks'
         ' statics/assets.'
)
@pass_cli_config
def install(cli_config, pre, dev, production):
    """Installs the  project locally.

    Installs dependencies, creates instance directory,
    links invenio.cfg + templates, copies images and other statics and finally
    builds front-end assets.
    """
    commands = InstallCommands(cli_config)
    flask_env = 'production' if production else 'development'
    steps = commands.install(pre=pre, dev=dev, flask_env=flask_env)
    on_fail = "Failed to install dependencies."
    on_success = "Dependencies installed successfully."

    run_steps(steps, on_fail, on_success)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 Northwestern University.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio-cli configuration file."""

from configparser import ConfigParser
from pathlib import Path

from ..errors import InvenioCLIConfigError
from .filesystem import get_created_files
from .process import ProcessResponse


class CLIConfig(object):
    """Invenio-cli configuration.

    It provides a combined interface to the local CLI configuration which
    is typically split between a
        .invenio file with general project configuration
        .invenio.private with per machine configuration
                         (not version controlled)
    """

    CONFIG_FILENAME = '.invenio'
    PRIVATE_CONFIG_FILENAME = '.invenio.private'
    CLI_SECTION = 'cli'
    COOKIECUTTER_SECTION = 'cookiecutter'
    FILES_SECTION = 'files'

    def __init__(self, project_dir='./', verbose=False):
        """Constructor.

        :param config_dir: Path to general cli config file.
        :param config_path: Absolute path to per machine cli config file.
        """
        self.config_path = Path(project_dir) / self.CONFIG_FILENAME
        self.config = ConfigParser()
        self.private_config_path = (
            Path(project_dir) / self.PRIVATE_CONFIG_FILENAME
        )
        self.private_config = ConfigParser()

        try:
            with open(self.config_path) as cfg_file:
                self.config.read_file(cfg_file)

            with open(self.private_config_path) as cfg_file:
                self.private_config.read_file(cfg_file)

        except FileNotFoundError as e:
            raise InvenioCLIConfigError(
                "Missing '{0}' file in current directory. "
                "Are you in the project folder?".format(e.filename),
            )

    def get_project_dir(self):
        """Returns path to project directory."""
        return Path(self.private_config[CLIConfig.CLI_SECTION]['project_dir'])

    def get_instance_path(self):
        """Returns path to application instance directory."""
        return Path(
            self.private_config[CLIConfig.CLI_SECTION]['instance_path']
        )

    def update_instance_path(self, new_instance_path):
        """Updates path to application instance directory."""
        self.private_config[CLIConfig.CLI_SECTION]['instance_path'] = \
            str(new_instance_path)

        with open(self.private_config_path, 'w') as configfile:
            self.private_config.write(configfile)

        return ProcessResponse(
            output=f"Instance path updated (new value {new_instance_path}).",
            status_code=0
        )

    def get_services_setup(self):
        """Returns bool whether services have been setup or not."""
        return self.private_config.getboolean(
            CLIConfig.CLI_SECTION, 'services_setup'
        )

    def update_services_setup(self, is_setup):
        """Updates path to application instance directory."""
        self.private_config[CLIConfig.CLI_SECTION]['services_setup'] = \
            str(is_setup)

        with open(self.private_config_path, 'w') as configfile:
            self.private_config.write(configfile)

        return ProcessResponse(
            output=f"Service setup status updated (new value {is_setup}).",
            status_code=0
        )

    def get_project_shortname(self):
        """Returns the project's shortname."""
        return self.config[CLIConfig.COOKIECUTTER_SECTION]['project_shortname']

    def get_db_type(self):
        """Returns the database type (mysql, postgresql)."""
        return self.config[CLIConfig.COOKIECUTTER_SECTION]['database']

    def get_es_version(self):
        """Returns the database type (mysql, postgresql)."""
        return self.config[CLIConfig.COOKIECUTTER_SECTION]['elasticsearch']

    @classmethod
    def write(cls, project_dir, flavour, replay, instance_path=None):
        """Write invenio-cli config file.

        :param project_dir: Folder to write the config file into
        :param flavour: 'RDM' or 'ILS'
        :param replay: dict of cookiecutter replay
        :return: absolute Path to config (project) directory
        """
        config_parser = ConfigParser()
        # Convert to absolute Path because simpler to reason about and pass
        project_dir = Path(project_dir).resolve()

        # Internal to Invenio-cli section
        config_parser[cls.CLI_SECTION] = {}
        config_parser[cls.CLI_SECTION]['flavour'] = flavour
        config_parser[cls.CLI_SECTION]['logfile'] = '/logs/invenio-cli.log'

        # Cookiecutter user input section
        config_parser[cls.COOKIECUTTER_SECTION] = {}
        for key, value in replay[cls.COOKIECUTTER_SECTION].items():
            config_parser[cls.COOKIECUTTER_SECTION][key] = value

        # Generated files section
        config_parser[cls.FILES_SECTION] = get_created_files(project_dir)

        config_path = project_dir / cls.CONFIG_FILENAME
        with open(config_path, 'w') as configfile:
            config_parser.write(configfile)

        # Internal to machine (not version controlled)
        config_parser = ConfigParser()
        config_parser[cls.CLI_SECTION] = {}
        config_parser[cls.CLI_SECTION]['project_dir'] = str(project_dir)
        config_parser[cls.CLI_SECTION]['instance_path'] = instance_path or ''
        config_parser[cls.CLI_SECTION]['services_setup'] = str(False)
        private_config_path = project_dir / cls.PRIVATE_CONFIG_FILENAME
        with open(private_config_path, 'w') as configfile:
            config_parser.write(configfile)

        return project_dir



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module docker_helper tests."""
import tempfile
from pathlib import Path

from invenio_cli.helpers.filesystem import get_created_files


def test_get_created_files():
    tmp_dir = tempfile.TemporaryDirectory()
    tmp_dir_path = Path(tmp_dir.name)
    (tmp_dir_path / 'foo').touch()
    (tmp_dir_path / 'bar').mkdir()
    (tmp_dir_path / 'bar' / 'baz').touch()

    files_tree = get_created_files(tmp_dir.name)

    expected_files_tree = {
        'foo': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855',  # noqa
        'bar': {
            'baz': 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'  # noqa
        }
    }
    assert files_tree == expected_files_tree



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 Northwestern University.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module commands/containers.py's tests."""

from pathlib import Path
from unittest.mock import call, patch

import pytest

from invenio_cli.commands import ContainersCommands


@pytest.mark.skip()
@pytest.fixture(scope='function')
def expected_setup_calls():
    return [
        call('project-shortname', 'invenio db init create'),
        call(
            'project-shortname',
            'invenio files location create --default default-location '
            '${INVENIO_INSTANCE_PATH}/data'
        ),
        call('project-shortname', 'invenio roles create admin'),
        call(
            'project-shortname',
            'invenio access allow superuser-access role admin'
        ),
        call('project-shortname', 'invenio index init'),
        # update_statics_and_assets call
        call('project-shortname', 'invenio collect'),
        call('project-shortname', 'invenio webpack create'),
        call('project-shortname', 'invenio webpack install --unsafe'),
        call('project-shortname', 'invenio webpack build')
    ]


@pytest.mark.skip()
@pytest.fixture(scope='function')
def expected_force_calls():
    return [
        call(
            'project-shortname',
            "invenio shell --no-term-title -c "
            "\"import redis; redis.StrictRedis.from_url(app.config['CACHE_REDIS_URL']).flushall(); print('Cache cleared')\""  # noqa
        ),
        call('project-shortname', 'invenio db destroy --yes-i-know'),
        call(
            'project-shortname',
            'invenio index destroy --force --yes-i-know'
        ),
        call('project-shortname', 'invenio index queue init purge')
    ]


@pytest.mark.skip()
@patch('invenio_cli.commands.containers.run_cmd')
@patch('invenio_cli.commands.containers.DockerHelper')
@patch('invenio_cli.helpers.process.popen')
def test_containerize_install(
        p_popen, p_docker_helper, p_run_cmd, mock_cli_config, mocked_pipe,
        expected_setup_calls):
    """Case: pre=False, force=False, install=True."""
    p_popen.return_value = mocked_pipe
    commands = ContainersCommands(mock_cli_config, p_docker_helper())
    commands.containerize(pre=False, force=False, install=True)

    commands.docker_helper.start_containers.assert_called()
    assert (
        commands.docker_helper.execute_cli_command.mock_calls ==
        expected_setup_calls
    )


@pytest.mark.skip()
@patch('invenio_cli.commands.containers.run_cmd')
@patch('invenio_cli.commands.containers.DockerHelper')
@patch('invenio_cli.helpers.process.popen')
def test_containerize_no_install(
        p_popen, p_docker_helper, p_run_cmd, mock_cli_config, mocked_pipe):
    """Case: pre=False, force=False, install=False"""
    p_popen.return_value = mocked_pipe
    commands = ContainersCommands(mock_cli_config, p_docker_helper())
    commands.docker_helper.execute_cli_command.reset_mock()

    commands.containerize(pre=False, force=False, install=False)

    assert (
        call('project-shortname', 'invenio webpack install --unsafe') not in
        commands.docker_helper.execute_cli_command.mock_calls
    )


@pytest.mark.skip()
@patch('invenio_cli.commands.containers.run_cmd')
@patch('invenio_cli.commands.containers.DockerHelper')
@patch('invenio_cli.helpers.process.popen')
def test_containerize_install_force(
        p_popen, p_docker_helper, p_run_cmd, mock_cli_config, mocked_pipe,
        expected_setup_calls, expected_force_calls):
    """Case: pre=False, force=True, install=True."""
    p_popen.return_value = mocked_pipe
    commands = ContainersCommands(mock_cli_config, p_docker_helper())
    commands.docker_helper.execute_cli_command.reset_mock()

    commands.containerize(pre=False, force=True, install=True)

    assert commands.docker_helper.execute_cli_command.mock_calls == (
        expected_force_calls + expected_setup_calls
    )


@pytest.mark.skip()
@patch('invenio_cli.commands.containers.run_cmd')
@patch('invenio_cli.commands.containers.DockerHelper')
@patch('invenio_cli.helpers.process.popen')
def test_containerize_install_pre_force(
        p_popen, p_docker_helper, p_run_cmd, mock_cli_config, mocked_pipe,
        expected_setup_calls, expected_force_calls):
    """Case: pre=True, force=True, install=True.
    Testing pre does not change the commands result.
    Only the locking of the Pipfile. No need to test all combinations.
    """
    p_popen.return_value = mocked_pipe
    commands = ContainersCommands(mock_cli_config, p_docker_helper())
    commands.docker_helper.execute_cli_command.reset_mock()
    commands.containerize(pre=True, force=True, install=True)

    assert commands.docker_helper.execute_cli_command.mock_calls == (
        expected_force_calls + expected_setup_calls
    )


@pytest.mark.skip()
@patch('invenio_cli.commands.containers.run_cmd')
@patch('invenio_cli.commands.containers.DockerHelper')
def test_update_statics_and_assets(
        p_docker_helper, p_run_cmd, mock_cli_config):
    commands = ContainersCommands(mock_cli_config, p_docker_helper())

    commands.update_statics_and_assets(install=True)

    expected_execute_cli_calls = [
        call('project-shortname', 'invenio collect'),
        call('project-shortname', 'invenio webpack create'),
        call('project-shortname', 'invenio webpack install --unsafe'),
        call('project-shortname', 'invenio webpack build')
    ]
    assert (
        commands.docker_helper.execute_cli_command.mock_calls ==
        expected_execute_cli_calls
    )

    # Second arg is str due to copy2 requirements
    expected_copy2_calls = [
        call(
            Path('project_dir/static/.'),
            '/opt/invenio/var/instance/static/'
        ),
        call(
            Path('project_dir/assets/.'),
            '/opt/invenio/var/instance/assets/'
        )
    ]
    assert commands.docker_helper.copy2.mock_calls == expected_copy2_calls

    # Reset for force=True assertions
    commands.docker_helper.execute_cli_command.reset_mock()

    commands.update_statics_and_assets(install=False)

    expected_execute_cli_calls = [
        call('project-shortname', 'invenio collect'),
        call('project-shortname', 'invenio webpack create'),
        call('project-shortname', 'invenio webpack build')
    ]
    assert (
        commands.docker_helper.execute_cli_command.mock_calls ==
        expected_execute_cli_calls
    )


@pytest.mark.skip()
@patch('invenio_cli.commands.containers.listdir', return_value=[])
@patch('invenio_cli.commands.containers.run_cmd')
@patch('invenio_cli.commands.containers.DockerHelper')
def test_lock_python_dependencies(
        p_docker_helper, p_run_cmd, p_listdir,
        mock_cli_config):
    commands = ContainersCommands(mock_cli_config, p_docker_helper())
    commands._lock_python_dependencies(pre=False)
    p_run_cmd.assert_called_with(['pipenv', 'lock'])

    commands._lock_python_dependencies(pre=True)

    p_run_cmd.assert_called_with(['pipenv', 'lock', '--pre'])


@pytest.mark.skip()
@patch('invenio_cli.commands.containers.listdir',
       return_value=['Pipfile.lock'])
@patch('invenio_cli.commands.containers.run_cmd')
@patch('invenio_cli.commands.containers.DockerHelper')
def test_no_lock_python_dependencies(
        p_docker_helper, p_run_cmd, p_listdir,
        mock_cli_config):
    commands = ContainersCommands(mock_cli_config, p_docker_helper())
    commands._lock_python_dependencies(pre=False)
    p_run_cmd.assert_not_called()

    commands._lock_python_dependencies(pre=True)
    p_run_cmd.assert_not_called()


@pytest.mark.skip()
@patch('invenio_cli.commands.containers.run_cmd')
@patch('invenio_cli.commands.containers.DockerHelper')
def test_demo(
        p_run_cmd, p_docker_helper, mock_cli_config):
    commands = ContainersCommands(mock_cli_config, p_docker_helper())

    commands.demo()

    assert (
        commands.docker_helper.execute_cli_command.mock_calls ==
        [call('project-shortname', 'invenio rdm-records demo')]
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""


import click

from ..commands import AssetsCommands, PackagesCommands
from .utils import pass_cli_config, run_steps


@click.group()
def packages():
    """Commands for package management."""


@packages.command()
@click.option('--pre', default=False, is_flag=True,
              help='Allows installation of alpha releases.')
@click.option('--dev', default=False, is_flag=True,
              help='Includes development dependencies.')
@pass_cli_config
def lock(cli_config, pre, dev):
    """Lock Python dependencies."""
    click.secho(
        f"Locking dependencies... Allow pre-releases: {pre}. " +
        f"Include dev-packages: {dev}.",
        fg="green"
    )
    steps = PackagesCommands.lock(pre, dev)
    on_fail = "Failed to lock dependencies."
    on_success = "Dependencies locked successfully."

    run_steps(steps, on_fail, on_success)


@packages.command()
@click.argument("packages", nargs=-1, type=str)
@click.option('-s', '--skip-build', default=False, is_flag=True,
              help='Do not rebuild the assets.')
@pass_cli_config
def install(cli_config, packages, skip_build):
    """Install one or a list of Python packages in the local environment."""
    if len(packages) < 1:
        raise click.UsageError("You must specify at least one package.")

    steps = PackagesCommands.install_packages(packages)

    on_fail = f"Failed to install packages {packages}."
    on_success = f"Packages {packages} installed successfully."

    run_steps(steps, on_fail, on_success)

    # FIXME: Migrate assets to steps.
    if not skip_build:
        click.secho("Rebuilding assets...")
        AssetsCommands(cli_config).update_statics_and_assets(
            force=True, flask_env='development')


@packages.command()
@pass_cli_config
def outdated(cli_config):
    """Show outdated Python dependencies."""
    steps = PackagesCommands.outdated_packages()

    on_fail = "Some of the packages need to be updated."
    on_success = "All packages are up to date."

    run_steps(steps, on_fail, on_success)


@packages.command()
@click.argument("version", required=False, type=str)
@pass_cli_config
def update(cli_config, version=None):
    """Update all or some Python python packages."""
    if version:
        db = cli_config.get_db_type()
        es = f"elasticsearch{cli_config.get_es_version()}"
        package = f"invenio-app-rdm[{db},{es}]~="
        steps = PackagesCommands.update_package_new_version(package, version)
        on_fail = f"Failed to update version {version}"
        on_success = f"Version {version} installed successfully."
    else:
        steps = PackagesCommands.update_packages()
        on_fail = "Failed to update packages."
        on_success = "Packages installed successfully."

    run_steps(steps, on_fail, on_success)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 Northwestern University.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module cookiecutter wrapper tests."""

from invenio_cli.helpers.cookiecutter_wrapper import CookiecutterWrapper


def test_constructor():
    cookiecutter = CookiecutterWrapper('RDM', ('', 'master'))

    assert cookiecutter.tmp_file is None
    assert cookiecutter.flavour == 'RDM'
    assert (
        cookiecutter.template == 'https://github.com/inveniosoftware/'
                                 'cookiecutter-invenio-rdm.git'
    )
    assert cookiecutter.template_name == 'cookiecutter-invenio-rdm'
    assert cookiecutter.checkout == 'master'


def test_extract_template_name_git_url():
    tpl_name = CookiecutterWrapper.extract_template_name(
        'https://github.com/inveniosoftware/cookiecutter-invenio-rdm.git'
    )

    assert tpl_name == 'cookiecutter-invenio-rdm'


def test_extract_template_name_path():
    tpl_name = CookiecutterWrapper.extract_template_name(
        '~/dev/cookiecutter-invenio-rdm/'
    )

    assert tpl_name == 'cookiecutter-invenio-rdm'



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 Northwestern University.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module docker_helper tests."""

from unittest.mock import patch

import pytest

from invenio_cli.helpers.docker_helper import DockerHelper


@pytest.mark.skip()
@patch('invenio_cli.helpers.docker_helper.run_cmd')
def test_start_containers(p_run_cmd):
    # Needed to fake call to docker-compose --version but not call
    # to docker-compose up
    def fake_normalize_name(self, project_shortname):
        return 'project_shortname'

    with patch.object(DockerHelper, '_normalize_name', fake_normalize_name):
        docker_helper = DockerHelper('project-shortname', local=True)

    docker_helper.start_containers()

    p_run_cmd.run.assert_called_with(
        ['docker-compose', '--file', 'docker-compose.yml', 'up',
         '--build', '-d']
    )

    with patch.object(DockerHelper, '_normalize_name', fake_normalize_name):
        docker_helper = DockerHelper('project-shortname', local=False)

    docker_helper.start_containers()

    p_run_cmd.run.assert_called_with(
        ['docker-compose', '--file', 'docker-compose.full.yml', 'up',
         '--build', '-d']
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio CLI environment helper."""

import contextlib
import os


@contextlib.contextmanager
def env(**environ):
    """Context manager to temporarily set environment variables.

    :param environ: Environment variables to set
    """
    old = {}
    for k, v in dict(environ).items():
        old[k] = os.environ.get(k, None)
        os.environ[k] = v
    try:
        yield
    finally:
        for k, v in old.items():
            if v is None:
                del os.environ[k]
            else:
                os.environ[k] = v



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019 CERN.
# Copyright (C) 2019 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio CLI Docker Compose class."""

import re

import docker

from .process import ProcessResponse, run_cmd, run_interactive

DOCKER_COMPOSE_VERSION_DASH = '1.21.0'


class DockerHelper(object):
    """Utility class to interact with docker-compose."""

    def __init__(self, project_shortname, local=True, log_config=None):
        """Constructor."""
        super(DockerHelper, self).__init__()
        self.container_prefix = self._normalize_name(project_shortname)
        self.local = local
        self.docker_client = docker.from_env()

    def _normalize_name(self, project_shortname):
        """Normalize the container name according to the compose version.

        Docker-Compose introduced support for dash and underscore in
        version 1.21.0.
        """
        dc_version_string = run_cmd(['docker-compose', '--version'])
        groups = re.search(r'1.[0-9]*.[0-9]*', dc_version_string.output)
        dc_version = groups.group(0)

        if dc_version < DOCKER_COMPOSE_VERSION_DASH:
            return re.sub(r'[^a-z0-9]', '', project_shortname)
        else:
            return project_shortname

    def build_images(self, pull=False, cache=True):
        """Build images.

        :param pull: Adds --pull to the docker-compose command.
        :param cache: Removes --no-cache to the docker-compose command.
        """
        command = [
            'docker-compose',
            '--file',
            'docker-compose.full.yml',
            'build',
        ]
        if pull:
            command.append('--pull')
        if not cache:
            command.append('--no-cache')

        # FIXME: To get real-time output
        return run_interactive(command)

    def start_containers(self, app_only=False):
        """Start containers according to the specified environment.

        :param app_only: Boot up only ui and api containers.
        """
        command = [
            'docker-compose',
            '--file',
            'docker-compose.yml' if self.local else 'docker-compose.full.yml',
            'up',
            '-d'  # --detach not supported in 1.17.0
        ]

        if app_only:
            command.extend(['web-ui', 'web-api'])

        return run_cmd(command)

    def stop_containers(self):
        """Stop currently running containers."""
        command = ['docker-compose', '--file', 'docker-compose.full.yml',
                   'stop']
        return run_cmd(command)

    def destroy_containers(self):
        """Stop and remove all containers, volumes and images."""
        command = ['docker-compose', '--file', 'docker-compose.full.yml',
                   'down', '--volumes']

        return run_cmd(command)

    def execute_cli_command(self, project_shortname, command):
        """Execute an invenio CLI command in the API container."""
        container_name = '{}_web-ui_1'.format(self.container_prefix)
        container = self.docker_client.containers.get(container_name)
        status = container.exec_run(
            cmd='/bin/bash -c "{}"'.format(command.replace('"', '\\"')),
            tty=True,
            stdout=True,
            stderr=True)
        # FIXME: What happens when exec_run fails? handle exception.
        return ProcessResponse(
            output=status.output.decode("utf-8").strip(),
            status_code=status.exit_code
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

from ..helpers.process import run_interactive


class FunctionStep(object):
    """A step which execution is a function call.

    Is composed of a function, arguments, and a message (feedback).
    """

    def __init__(self, func, args=None, message=None):
        """Constructor."""
        self.func = func
        self.args = args or {}
        self.message = message

    def execute(self):
        """Execute the function with the given arguments."""
        return self.func(**self.args)


class CommandStep(object):
    """A step which execution is a command run.

    Is composed of a command, an environment, and a message (feedback).
    """

    def __init__(self, cmd, env=None, message=None, skippable=False):
        """Constructor."""
        self.cmd = cmd
        self.env = env
        self.message = message
        self.skippable = skippable

    def execute(self):
        """Execute the function with the given arguments."""
        return run_interactive(self.cmd, self.env, self.skippable)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 Northwestern University.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Commands tests fixtures."""

from pathlib import Path
from unittest.mock import Mock

import pytest


@pytest.fixture
def mock_cli_config():
    """Mock CLIConfig object."""
    class MockCLIConfig(object):
        def __init__(self):
            self.services_setup = False

        def get_project_dir(self):
            return Path('project_dir')

        def get_instance_path(self):
            return Path('instance_dir')

        def get_services_setup(self):
            return self.services_setup

        def update_services_setup(self, is_setup):
            self.services_setup = bool(is_setup)

        def get_project_shortname(self):
            return 'project-shortname'

        def get_db_type(self):
            return 'postgresql'

    return MockCLIConfig()


@pytest.fixture(scope="function")
def mocked_pipe():
    """Mock success return subprocess pipe for popen."""
    mocked_pipe = Mock()
    attrs = {
        'communicate.return_value': (b'output', b'error'),
        'returncode': 0
    }
    mocked_pipe.configure_mock(**attrs)
    return mocked_pipe


@pytest.fixture()
def testpkg():
    """Test package.json file."""
    return str(Path(__file__).parent / 'testpkg')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

from .assets import AssetsCommands
from .commands import Commands
from .containers import ContainersCommands
from .install import InstallCommands
from .local import LocalCommands
from .packages import PackagesCommands
from .requirements import RequirementsCommands
from .services import ServicesCommands
from .upgrade import UpgradeCommands

__all__ = (
    "AssetsCommands",
    "Commands",
    "ContainersCommands",
    "InstallCommands",
    "LocalCommands",
    "PackagesCommands",
    "RequirementsCommands",
    "ServicesCommands",
    "UpgradeCommands"
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""


import subprocess
from functools import partial
from pathlib import Path

import click
from pynpm import NPMPackage

from ..helpers import env
from ..helpers.process import ProcessResponse, run_interactive
from .local import LocalCommands
from .steps import CommandStep, FunctionStep


class AssetsCommands(LocalCommands):
    """Local installation commands."""

    def __init__(self, cli_config):
        """Constructor."""
        super(AssetsCommands, self).__init__(cli_config)

    @staticmethod
    def _module_pkg(path):
        """NPM package for the given path."""
        return NPMPackage(Path(path) / 'package.json')

    def _assets_pkg(self):
        """NPM package for the instance's webpack project."""
        return self._module_pkg(self.cli_config.get_instance_path() / 'assets')

    @staticmethod
    def _watch_js_module(pkg):
        """Watch the JS module for changes."""
        click.secho('Starting watching module...', fg='green')
        status_code = pkg.run_script('watch')
        if status_code == 0:
            return ProcessResponse(
                output="Watched module successfully.",
                status_code=0
            )
        else:
            return ProcessResponse(
                error=f"Unable to set watcher. Got status code {status_code}",
                status_code=status_code
            )

    @staticmethod
    def _run_script(module_pkg):
        """Run script and return a ProcessResponse."""
        status_code = module_pkg.run_script("link-dist")
        if status_code == 0:
            return ProcessResponse(
                output="Module linked correctly to global",
                status_code=0
            )
        else:
            return ProcessResponse(
                error=f"Unable to link-dist. Got error code {status_code}",
                status_code=status_code
            )

    @staticmethod
    def _npm_install_command(path):
        """Run command and return a ProcessResponse."""
        status_code = subprocess.call(["npm", "install", "--prefix", path])
        if status_code == 0:
            return ProcessResponse(
                output="Dependent packages installed correctly",
                status_code=0
            )
        else:
            return ProcessResponse(
                error=f"Unable to install dependent packages. "
                      "Got error code {status_code}",
                status_code=status_code
            )

    @staticmethod
    def _build_script(module_pkg):
        """Run script and return a ProcessResponse."""
        status_code = module_pkg.run_script("build")
        if status_code == 0:
            return ProcessResponse(
                output="Built correctly",
                status_code=0
            )
        else:
            return ProcessResponse(
                error=f"Unable to build. Got error code {status_code}",
                status_code=status_code
            )

    @staticmethod
    def _assets_link(assets_pkg, module_pkg):
        try:
            module_name = module_pkg.package_json['name']
        except FileNotFoundError as e:
            return ProcessResponse(
                error="No module found on the specified path. "
                      f"File not found {e.filename}",
                status_code=1
            )

        status_code = assets_pkg.link(module_name)
        if status_code == 0:
            return ProcessResponse(
                output="Global module linked correctly to local folder",
                status_code=0
            )
        else:
            return ProcessResponse(
                error=f"Unable to link module. Got error code {status_code}",
                status_code=status_code
            )

    def watch_assets(self):
        """High-level command to watch assets for changes."""
        # Commands
        prefix = ['pipenv', 'run']
        watch_cmd = prefix + ['invenio', 'webpack', 'run', 'start']

        with env(FLASK_ENV='development'):
            # Collect into statics/ and assets/ folder
            click.secho('Starting assets watching (press CTRL+C to stop)...',
                        fg='green')
            run_interactive(watch_cmd, env={'PIPENV_VERBOSITY': "-1"})

    def link_js_module(self, path):
        """High-level command to install and build a JS module."""
        module_pkg = self._module_pkg(path)
        assets_pkg = self._assets_pkg()

        steps = [
            FunctionStep(  # Install dependent packages
                func=self._npm_install_command,
                args={"path": path},
                message="Installing dependent packages..."
            ),
            FunctionStep(  # Run build script
                func=self._build_script,
                args={"module_pkg": module_pkg},
                message="Building..."
            ),
            FunctionStep(  # Create link to global folder
                func=self._run_script,
                args={"module_pkg": module_pkg},
                message="Linking module to global dist..."
            ),
            FunctionStep(  # Link the global folder to the assets folder.
                func=self._assets_link,
                args={
                    "assets_pkg": assets_pkg,
                    "module_pkg": module_pkg
                },
                message="Linking module to assets..."
            )
        ]

        return steps

    def watch_js_module(self, path, link=True):
        """High-level command to watch a JS module for changes."""
        steps = []
        if link:
            steps.extend(self.link_js_module(path))

        steps.append(
            FunctionStep(
                func=self._watch_js_module,
                args={"pkg": self._module_pkg(path)},
                message="Starting watching module..."
            )
        )

        return steps



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

#####
# IMPORTANT NOTE: If you are going to modify any code here.
# Check `docker-service-cli` since the original code belongs there,
# and any bug might have already been fixed there.
# The reason for the copy-paste was to simplify the complexity of the
# integration. Might be integrated in the future when `docker-services-cli`
# reaches a higher maturity level.
#####

import time

import click

from ..helpers.process import run_cmd


class ServicesHealthCommands(object):
    """Services status commands."""

    @classmethod
    def es_healthcheck(cls, *args, **kwargs):
        """Elasticsearch healthcheck."""
        verbose = kwargs['verbose']

        return run_cmd([
            "curl", "-f",
            "localhost:9200/_cluster/health?wait_for_status=yellow"
        ])

    @classmethod
    def postgresql_healthcheck(cls, *args, **kwargs):
        """Postgresql healthcheck."""
        filepath = kwargs['filepath']
        verbose = kwargs['verbose']

        return run_cmd([
            "docker-compose", "--file", filepath,
            "exec", "-T", "db", "bash", "-c", "pg_isready",
        ])

    @classmethod
    def mysql_healthcheck(cls, *args, **kwargs):
        """Mysql healthcheck."""
        filepath = kwargs['filepath']
        verbose = kwargs['verbose']
        password = kwargs['project_shortname']

        return run_cmd([
            "docker-compose", "--file", filepath,
            "exec", "-T", "db", "bash", "-c",
            f"mysql -p{password} -e \"select Version();\"",
        ])

    @classmethod
    def redis_healthcheck(cls, *args, **kwargs):
        """Redis healthcheck."""
        filepath = kwargs['filepath']
        verbose = kwargs['verbose']

        return run_cmd([
            "docker-compose", "--file", filepath,
            "exec", "-T", "cache", "bash", "-c",
            "redis-cli ping", "|", "grep 'PONG'", "&>/dev/null;",
        ])

    @classmethod
    def wait_for_services(
        cls,
        services,
        project_shortname,
        filepath="docker-services.yml",
        max_retries=6,
        verbose=False,
    ):
        """Wait for services to be up.

        It performs configured healthchecks in a serial fashion, following the
        order given in the ``up`` command. If the services is an empty list,
        to be compliant with `docker-compose` it will perform the healthchecks
        of all the services.
        """
        if len(services) == 0:
            services = HEALTHCHECKS.keys()

        for service in services:
            exp_backoff_time = 2
            try_ = 1
            check = HEALTHCHECKS[service]
            ready = check(
                filepath=filepath,
                verbose=verbose,
                project_shortname=project_shortname,
            )
            while not ready and try_ < max_retries:
                click.secho(
                    f"{service} not ready at {try_} retries, waiting " +
                    f"{exp_backoff_time}s",
                    fg="yellow"
                )
                try_ += 1
                time.sleep(exp_backoff_time)
                exp_backoff_time *= 2
                ready = check(
                    filepath=filepath,
                    verbose=verbose,
                    project_shortname=project_shortname,
                ).status_code == 0

            if not ready:
                click.secho(f"Unable to boot up {service}", fg="red")
                exit(1)
            else:
                click.secho(f"{service} up and running!", fg="green")


HEALTHCHECKS = {
    "es": ServicesHealthCommands.es_healthcheck,
    "postgresql": ServicesHealthCommands.postgresql_healthcheck,
    "mysql": ServicesHealthCommands.mysql_healthcheck,
    "redis": ServicesHealthCommands.redis_healthcheck
}
"""Health check functions module path, as string."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

from ..helpers.docker_helper import DockerHelper
from ..helpers.process import ProcessResponse
from .commands import Commands
from .services_health import HEALTHCHECKS, ServicesHealthCommands
from .steps import CommandStep, FunctionStep


class ServicesCommands(Commands):
    """Service CLI commands."""

    def __init__(self, cli_config, docker_helper=None):
        """Constructor."""
        super(ServicesCommands, self).__init__(cli_config)
        self.docker_helper = docker_helper or \
            DockerHelper(cli_config.get_project_shortname(), local=True)

    def ensure_containers_running(self):
        """Ensures containers are running."""
        project_shortname = self.cli_config.get_project_shortname()

        self.docker_helper.start_containers()

        ServicesHealthCommands.wait_for_services(
            services=["redis", self.cli_config.get_db_type(), "es"],
            project_shortname=project_shortname,
        )
        return ProcessResponse(
            output="Containers started and healthy.",
            status_code=0,
        )

    def services_expected_status(self, expected):
        """Checks if the services have the expected status."""
        if not self.cli_config.get_services_setup() == expected:
            return ProcessResponse(
                error="Services status inconsistent." +
                      f"Expected {expected} obtained {not expected}",
                status_code=1
            )

        return ProcessResponse(
                output=f"Services setup status consistent.",
                status_code=0
            )

    def _cleanup(self):
        """Services cleanup steps."""
        steps = [
            CommandStep(cmd=[
                'pipenv', 'run', 'invenio', 'shell', '--no-term-title', '-c',
                "import redis; redis.StrictRedis.from_url(app.config['CACHE_REDIS_URL']).flushall(); print('Cache cleared')"],  # noqa
                env={'PIPENV_VERBOSITY': "-1"},
                message="Flushing Redis...",
                skippable=True
            ),
            CommandStep(
                cmd=['pipenv', 'run', 'invenio', 'db', 'destroy',
                     '--yes-i-know'],
                env={'PIPENV_VERBOSITY': "-1"},
                message="Destroying database...",
                skippable=True
            ),
            CommandStep(
                cmd=['pipenv', 'run', 'invenio', 'index', 'destroy',
                     '--force', '--yes-i-know'],
                env={'PIPENV_VERBOSITY': "-1"},
                message="Destroying indices...",
                skippable=True
            ),
            CommandStep(
                cmd=['pipenv', 'run', 'invenio', 'index', 'queue',
                     'init', 'purge'],
                env={'PIPENV_VERBOSITY': "-1"},
                message="Purging queues...",
                skippable=True
            ),
            FunctionStep(func=self.cli_config.update_services_setup,
                args={"is_setup": False},
                message="Updating service setup status (False)..."
            )
        ]

        return steps

    def _setup(self):
        """Services initialization steps."""
        steps = [
            FunctionStep(
                func=self.services_expected_status,
                args={"expected": False},
                message="Checking services are not setup..."
            ),
            CommandStep(
                cmd=['pipenv', 'run', 'invenio', 'db', 'init', 'create'],
                env={'PIPENV_VERBOSITY': "-1"},
                message="Creating database..."
            ),
            CommandStep(
                cmd=['pipenv', 'run', 'invenio', 'files', 'location',
                     'create', '--default', 'default-location',
                     "{}/data".format(self.cli_config.get_instance_path())],
                env={'PIPENV_VERBOSITY': "-1"},
                message="Creating files location..."
            ),
            CommandStep(
                cmd=['pipenv', 'run', 'invenio', 'roles', 'create', 'admin'],
                env={'PIPENV_VERBOSITY': "-1"},
                message="Creating admin role..."
            ),
            CommandStep(
                cmd=['pipenv', 'run', 'invenio', 'access', 'allow',
                     'superuser-access', 'role', 'admin'],
                env={'PIPENV_VERBOSITY': "-1"},
                message="Allowing superuser access to admin role..."
            ),
            CommandStep(
                cmd=['pipenv', 'run', 'invenio', 'index', 'init'],
                env={'PIPENV_VERBOSITY': "-1"},
                message="Creating indices..."
            ),
            FunctionStep(
                func=self.cli_config.update_services_setup,
                args={"is_setup": True},
                message="Updating service setup status (True)..."
            )
        ]

        return steps

    def demo(self):
        """Steps to add demo records into the instance."""
        steps = [
            CommandStep(
                cmd=['pipenv', 'run', 'invenio', 'rdm-records', 'demo'],
                env={'PIPENV_VERBOSITY': "-1"},
                message="Creating demo records..."
            )
        ]

        return steps

    def fixtures(self):
        """Steps to set up the required fixtures for the instance."""
        command = ['pipenv', 'run', 'invenio', 'rdm-records', 'fixtures']
        steps = [
            CommandStep(
                cmd=command,
                env={'PIPENV_VERBOSITY': "-1"},
                message="Creating fixtures..."
            )
        ]

        return steps

    def setup(self, force, demo_data=True, stop=False, services=True):
        """Steps to setup services' containers.

        A check in invenio-cli's config file is done to see if one-time setup
        has been executed before.
        """
        steps = []

        if services:
            steps.append(
                FunctionStep(func=self.ensure_containers_running,
                             message="Making sure containers are up...")
            )
        if force:
            steps.extend(self._cleanup())

        steps.extend(self._setup())
        steps.extend(self.fixtures())

        if demo_data:
            steps.extend(self.demo())

        if stop:
            steps.append(
                FunctionStep(
                    func=self.docker_helper.stop_containers,
                    message="Stopping containers...."
                )
            )
        return steps

    def start(self):
        """Steps to start services' containers."""
        steps = [
            FunctionStep(func=self.ensure_containers_running,
                         message="Making sure containers are up...")
        ]

        return steps

    def stop(self):
        """Stops containers."""
        steps = [
            FunctionStep(
                func=self.docker_helper.stop_containers,
                message="Stopping containers..."
            )
        ]

        return steps

    def destroy(self):
        """Steps to destroy the services's containers."""
        steps = [
            FunctionStep(
                func=self.docker_helper.destroy_containers,
                message="Destroying containers..."
            ),
            FunctionStep(
                func=self.cli_config.update_services_setup,
                args={"is_setup": False},
                message="Updating service setup status (False)..."
            )
        ]

        return steps

    def status(self, services, verbose):
        """Checks the status of the given service.

        :returns: A list of the same length than services. Each item will be a
                  code corresponding to: 0 success, 1 failure, 2 healthcheck
                  not defined.
        """
        project_shortname = self.cli_config.get_project_shortname()
        statuses = []
        for service in services:
            check = HEALTHCHECKS.get(service)
            if check:
                result = check(
                    filepath="docker-services.yml",
                    verbose=verbose,
                    project_shortname=project_shortname,
                )
                # Append 0 if OK, else 1
                # FIXME: Deal with codes higher than 1. Needed?
                code = 0 if result.status_code == 0 else 1
                statuses.append(code)
            else:
                statuses.append(2)

        return statuses



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2021 TU Wien.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

import re
import sys
from os import listdir

from ..helpers.process import ProcessResponse, run_cmd, run_interactive
from .steps import FunctionStep


class RequirementsCommands(object):
    """Pre-requirements check."""

    @classmethod
    def _version_from_string(cls, string):
        """Gets the version from a given string."""
        groups = re.search(r'[0-9]*\.[0-9]*\.[0-9]*', string)
        return groups.group(0)

    @classmethod
    def _check_version(cls, binary, version,
                       major, minor=-1, patch=-1, exact=False):
        """Checks a version."""
        parts = version.split('.')

        if len(parts) != 3:
            return ProcessResponse(
                error=f"{binary} incorrect version format or not found. "
                      "Check that it is installed correctly",
                status_code=1
            )

        parts = [int(num) for num in parts]

        if (exact and parts[0] != major) or parts[0] < major:
            return ProcessResponse(
                error=f"{binary} major version missmatch. "
                      f"Got {parts[0]} expected {major}",
                status_code=1
            )
        if (exact and parts[1] != minor and minor != -1) or parts[1] < minor:
            return ProcessResponse(
                error=f"{binary} minor version missmatch. "
                      f"Got {parts[1]} expected {minor}",
                status_code=1
            )
        if (exact and parts[2] != patch and patch != -1) or parts[2] < patch:
            return ProcessResponse(
                error=f"{binary} patch version missmatch. "
                      f"Got {parts[2]} expected {patch}",
                status_code=1
            )

        return ProcessResponse(
                output=f"{binary} version OK. Got {version}.",
                status_code=0
            )

    @classmethod
    def check_node_version(cls, major, minor=-1, patch=-1, exact=False):
        """Check the node version."""
        # Output comes in the form of 'v14.4.0\n'
        result = run_cmd(["node", "--version"])
        version = cls._version_from_string(result.output.strip())
        return cls._check_version("Node", version, major, minor, patch, exact)

    @classmethod
    def check_python_version(cls, major, minor=-1, patch=-1, exact=False):
        """Check the python version."""
        # check the version of the currently executed Python, as
        # 'invenio-cli' will create a virtualenv with the stated Python
        # version (via pipenv) anyway
        version_info = sys.version_info
        version = "{}.{}.{}".format(
            version_info.major,
            version_info.minor,
            version_info.micro,
        )

        return cls._check_version(
            "Python", version, major, minor, patch, exact)

    @classmethod
    def check_docker_version(cls, major, minor=-1, patch=-1, exact=False):
        """Check the docker version."""
        # Output comes in the form of
        # 'Docker version 19.03.13, build 4484c46d9d\n'
        try:
            result = run_cmd(["docker", "--version"])
            version = cls._version_from_string(result.output.strip())
            return cls._check_version(
                "Docker", version, major, minor, patch, exact)
        except Exception as err:
            return ProcessResponse(
                error=f"Docker not found. Got {err}.", status_code=1)

    @classmethod
    def check_docker_compose_version(cls, major, minor=-1, patch=-1,
                                     exact=False):
        """Check the docker compose version."""
        # Output comes in the form of
        # 'docker-compose version 1.27.4, build 4484c46d9d\n'
        try:
            result = run_cmd(["docker-compose", "--version"])
            version = cls._version_from_string(result.output.strip())
            return cls._check_version(
                "Docker Compose", version, major, minor, patch, exact
            )
        except Exception as err:
            return ProcessResponse(
                error=f"Docker Compose not found. Got {err}.", status_code=1
            )

    @classmethod
    def check_imagemagick_version(cls, major, minor=-1, patch=-1,
                                  exact=False):
        """Check the ImageMagick version."""
        # Output comes in the form of 'ImageMagick, version 7.0.11-13\n'
        try:
            result = run_cmd(["convert", "--version"])
            version = cls._version_from_string(result.output.strip())
            return cls._check_version(
                "ImageMagick", version, major, minor, patch, exact
            )
        except Exception as err:
            return ProcessResponse(
                error=f"ImageMagick not found. Got {err}.",
                status_code=1,
            )

    @classmethod
    def check_pipenv_installed(cls):
        """Check the pipenv version."""
        # Output comes in the form of 'pipenv, version 2020.11.15\n'
        result = run_cmd(["pipenv", "--version"])

        parts = result.output.strip().split(',')
        if parts[0] != 'pipenv':
            return ProcessResponse(
                error=f"Pipenv not found. Got {result.error}.",
                status_code=1
            )

        version = cls._version_from_string(parts[1])

        return ProcessResponse(
                output=f"Pipenv OK. Got version {version}.",
                status_code=0
            )

    @classmethod
    def check_dev(cls):
        """Steps to check the development pre-requisites."""
        steps = [
            FunctionStep(
                func=cls.check_node_version,
                args={"major": 14, "exact": True},
                message="Checking Node version..."
            )
        ]

        return steps

    @classmethod
    def check(cls, development=False):
        """Steps to check the pre-requisites."""
        steps = [
            FunctionStep(
                func=cls.check_python_version,
                args={"major": 3, "minor": 6},
                message="Checking Python version..."
            ),
            FunctionStep(
                func=cls.check_pipenv_installed,
                message="Checking Pipenv is installed..."
            ),
            FunctionStep(
                func=cls.check_docker_version,
                args={"major": 0, "minor": 0},
                message="Checking Docker version..."
            ),
            FunctionStep(
                func=cls.check_docker_compose_version,
                args={"major": 1, "minor": 17},
                message="Checking Docker Compose version..."
            ),
            FunctionStep(
                func=cls.check_imagemagick_version,
                args={"major": 0, "minor": 0},
                message="Checking ImageMagick version..."
            )
        ]

        if development:
            steps.extend(cls.check_dev())

        return steps



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

from .cli import invenio_cli

__all__ = (
    "invenio_cli",
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019 CERN.
# Copyright (C) 2019 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Filesystem helper functions."""

import errno
import hashlib
from os import listdir, remove, symlink
from os.path import isdir
from pathlib import Path

from .process import ProcessResponse

# Defining buffer size to avoid using too much memory
BUF_SIZE = 65536  # 64 KB


def hash_file(path_to_file):
    """Hash file to check for consistency."""
    sha256 = hashlib.sha256()

    with open(path_to_file, 'rb') as f:
        while True:
            data = f.read(BUF_SIZE)
            if not data:
                break
            sha256.update(data)

    return sha256.hexdigest()


def get_created_files(folder):
    """Return the generated tree of files (and their hash) and folders."""
    files = {}
    for name in listdir(folder):
        path = Path(folder)  # Current path

        # Add files and their hash
        if not isdir(path / name):
            files[name] = hash_file(path / name)
        # Add dirs and their files
        else:
            files[name] = get_created_files(path / name)

    return files


def force_symlink(target, link_name):
    """Forcefully create symlink at link_name pointing to target."""
    output = f"Symlinked {target} successfully."
    try:
        symlink(target, link_name)
    except OSError as e:
        if e.errno == errno.EEXIST:
            remove(link_name)
            symlink(target, link_name)
            output = output + "Deleted already existing link."

    return ProcessResponse(
        output=output,
        status_code=0
    )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019 CERN.
# Copyright (C) 2019 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio CLI Process helper module."""

from os import environ
from subprocess import PIPE, CalledProcessError
from subprocess import Popen as popen
from subprocess import run


class ProcessResponse():
    """Process response class."""

    def __init__(self, output=None, error=None, status_code=0, warning=False):
        """Constructor.

        By default it is a successful response (0) wiht no error nor ouput.
        """
        self.output = output
        self.error = error
        self.status_code = status_code
        self.warning = warning


def run_cmd(command):
    """Runs a given command and returns a ProcessResponse."""
    p = popen(command, stdout=PIPE, stderr=PIPE)
    output, error = p.communicate()
    output = output.decode("utf-8")
    error = error.decode("utf-8")

    return ProcessResponse(output, error, p.returncode)


def run_interactive(command, env=None, skippable=False):
    """Runs a given command without blocking, allows interactive shells.

    Stdout and stderr are not piped and therefore allows interaction.
    :param command: The command to run, in array form.
    :param env: A dict of variables to add to the environment.
    """
    full_env = environ.copy()  # Need to inherit the global one
    if env:
        for var, val in env.items():
            full_env[var] = val

    try:
        response = run(command, check=True, env=full_env)
        return ProcessResponse(
            output=None, error=None, status_code=0)
    except CalledProcessError as e:
        if skippable:
            return ProcessResponse(
                output=e.stdout, error=e.stderr, status_code=0, warning=True)
        else:
            return ProcessResponse(
                output=e.stdout, error=e.stderr, status_code=e.returncode)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019-2021 CERN.
# Copyright (C) 2019-2021 Northwestern University.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio Cookiecutter Config class."""

import json
import tempfile
from pathlib import Path

import yaml
from cookiecutter import replay
from cookiecutter.config import DEFAULT_CONFIG
from cookiecutter.exceptions import OutputDirExistsException
from cookiecutter.main import cookiecutter


class CookiecutterWrapper(object):
    """Cookiecutter helper object for InvenioCLI."""

    @classmethod
    def extract_template_name(cls, template):
        """Extract template name from a template URL."""
        name = template.rstrip('/').rpartition("/")[2]
        git = '.git'
        if name.endswith(git):
            name = name[:-len(git)]
        return name

    def __init__(self, flavour, template_checkout):
        """Constructor.

        :param flavour: "RDM" or something else: String
        :param cookiecutter: tuple of (template: URL, checkout: String)
        """
        self.tmp_file = None
        self.template_name = None
        self.flavour = flavour

        if self.flavour.upper() == 'RDM':
            self.template = (
                template_checkout[0] or
                'https://github.com/inveniosoftware/'
                'cookiecutter-invenio-rdm.git'
            )
            self.template_name = self.extract_template_name(self.template)
            self.checkout = template_checkout[1] or 'v3.0'

    def cookiecutter(self):
        """Wrap cookiecutter call."""
        return cookiecutter(
            config_file=self.create_and_dump_config_file(),
            **self.repository()
        )

    def repository(self):
        """Get the cookiecutter repository options."""
        return {
            'template': self.template,
            # NOTE: if template is not a git url, then checkout is ignored
            'checkout': self.checkout
        }

    def create_and_dump_config_file(self):
        """Create a tmp file to store used configuration."""
        if not self.tmp_file:
            self.tmp_file = tempfile.NamedTemporaryFile(mode='w+')

        config = DEFAULT_CONFIG.copy()
        # BUG: when dumping {}, it's read back as a string '{}'
        config['default_context'] = None
        config['replay_dir'] = tempfile.gettempdir()

        yaml.dump(config, self.tmp_file)

        return self.tmp_file.name

    def remove_config(self):
        """Remove the tmp file."""
        if self.tmp_file:
            self.tmp_file.close()

    def get_replay(self):
        """Retrieve dict of user input values."""
        if self.template_name:
            return replay.load(tempfile.gettempdir(), self.template_name)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

from ..helpers.env import env
from ..helpers.process import run_interactive
from .steps import CommandStep


class Commands(object):
    """Abstraction over CLI commands that are either local or containerized."""

    def __init__(self, cli_config):
        """Constructor.

        :param cli_config: :class:CLIConfig instance
        """
        self.cli_config = cli_config

    @classmethod
    def shell(cls):
        """Start a shell in the virtual environment."""
        command = ['pipenv', 'shell', ]
        return run_interactive(command, env={'PIPENV_VERBOSITY': "-1"})

    @classmethod
    def pyshell(cls, debug=False):
        """Start a Python shell."""
        with env(FLASK_ENV='development' if debug else 'production'):
            command = ['pipenv', 'run', 'invenio', 'shell']
            return run_interactive(command, env={'PIPENV_VERBOSITY': "-1"})

    def destroy(self):
        """Destroys the instance's virtualenv.

        NOTE: This function has no knowledge of the existence of services.
              Refer to services.py to destroy services' containers.
        """
        steps = [
            CommandStep(
                cmd=['pipenv', '--rm'],
                env={'PIPENV_VERBOSITY': "-1"},
                message="Destroying virtual environment"
            )
        ]

        return steps



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

from ..helpers.docker_helper import DockerHelper
from .packages import PackagesCommands
from .services import ServicesCommands
from .services_health import HEALTHCHECKS
from .steps import FunctionStep


class ContainersCommands(ServicesCommands):
    """Containerized environment CLI commands."""

    def __init__(self, cli_config, docker_helper=None):
        """Constructor."""
        docker_helper = docker_helper or \
            DockerHelper(cli_config.get_project_shortname(), local=False)

        super(ContainersCommands, self).__init__(cli_config, docker_helper)

    def build(self, pull=True, cache=True):
        """Return the steps to build images.

        :param pull: Attempt to pull newer versions of the images.
        :param cache: Use cached images and layers.
        """
        steps = [
            FunctionStep(
                func=PackagesCommands.is_locked,
                message="Checking if dependencies are locked."
            ),
            FunctionStep(
                func=self.docker_helper.build_images,
                args={"pull": pull, "cache": cache},
                message="Building images..."
            )

        ]

        return steps

    def _cleanup(self, project_shortname="/opt/var/instance/"):
        """Steps to cleanup commands."""
        steps = [
            FunctionStep(
                func=self.docker_helper.execute_cli_command,
                args={
                    "project_shortname": project_shortname,
                    "command": "invenio shell --no-term-title -c "
                               "\"import redis; redis.StrictRedis.from_url(app.config['CACHE_REDIS_URL']).flushall(); print('Cache cleared')\""  # noqa
                },
                message="Flushing redis cache..."
            ),
            FunctionStep(
                func=self.docker_helper.execute_cli_command,
                args={
                    "project_shortname": project_shortname,
                    "command": "invenio db destroy --yes-i-know"
                },
                message="Deleting database..."
            ),
            FunctionStep(
                func=self.docker_helper.execute_cli_command,
                args={
                    "project_shortname": project_shortname,
                    "command": "invenio index destroy --force --yes-i-know"
                },
                message="Deleting indices..."
            ),
            FunctionStep(
                func=self.docker_helper.execute_cli_command,
                args={
                    "project_shortname": project_shortname,
                    "command": "invenio index queue init purge"
                },
                message="Purging queues..."
            ),
            FunctionStep(func=self.cli_config.update_services_setup,
                args={"is_setup": False},
                message="Updating service setup status (False)..."
            )
        ]

        return steps

    def _setup(self, project_shortname="/opt/var/instance/"):
        """Steps to initialize services."""
        steps = [
            FunctionStep(
                func=self.docker_helper.execute_cli_command,
                args={
                    "project_shortname": project_shortname,
                    "command": "invenio db init create"
                },
                message="Creating database..."
            ),
            FunctionStep(
                func=self.docker_helper.execute_cli_command,
                args={
                    "project_shortname": project_shortname,
                    "command": "invenio files location create --default "
                               "default-location "
                               "${INVENIO_INSTANCE_PATH}/data"
                },
                message="Creating files location..."
            ),
            FunctionStep(
                func=self.docker_helper.execute_cli_command,
                args={
                    "project_shortname": project_shortname,
                    "command": "invenio roles create admin"
                },
                message="Creating admin role..."
            ),
            FunctionStep(
                func=self.docker_helper.execute_cli_command,
                args={
                    "project_shortname": project_shortname,
                    "command": "invenio access allow "
                               "superuser-access role admin"
                },
                message="Assigning superuser access to admin role..."
            ),
            FunctionStep(
                func=self.docker_helper.execute_cli_command,
                args={
                    "project_shortname": project_shortname,
                    "command": "invenio index init"
                },
                message="Creating indices..."
            ),
            FunctionStep(
                func=self.cli_config.update_services_setup,
                args={"is_setup": True},
                message="Updating service setup status (True)..."
            )
        ]

        return steps

    def demo(self, project_shortname):
        """Steps to demo records into the instance."""
        steps = [
            FunctionStep(
                func=self.docker_helper.execute_cli_command,
                args={
                    "project_shortname": project_shortname,
                    "command": "invenio rdm-records demo"
                },
                message="Creating demo records..."
            )
        ]

        return steps

    def fixtures(self, project_shortname):
        """Steps to set up the required fixtures for the instance."""
        steps = [
            FunctionStep(
                func=self.docker_helper.execute_cli_command,
                args={
                    "project_shortname": project_shortname,
                    "command": "invenio rdm-records fixtures"
                },
                message="Creating fixtures..."
            )
        ]

        return steps

    def setup(self, force, demo_data=True, stop=False, services=True):
        """Return the steps to setup containerize services.

        :param force: Remove existing content (db, indices, etc.).
        :param demo_data: Include demo records.
        :param stop: Stop services after setup.
        """
        steps = []

        if services:
            steps.append(
                FunctionStep(
                    func=self.ensure_containers_running,
                    message="Making sure containers are up..."
                )
            )

        project_shortname = self.cli_config.get_project_shortname()

        if force:
            steps.extend(self._cleanup(project_shortname))

        steps.extend(self._setup(project_shortname))
        steps.extend(self.fixtures(project_shortname))

        if demo_data:
            steps.extend(self.demo(project_shortname))

        if stop:
            steps.append(
                FunctionStep(
                    func=self.docker_helper.stop_containers,
                    message="Stopping containers...."
                )
            )

        return steps

    def start(self, lock=False, build=False, setup=False, demo_data=True,
              services=True):
        """Return the steps to start service and application containers.

        :param lock: Lock dependencies.
        :param build: Build containers if not built.
        :param setup: Setup services (db, indices, etc.).
        :param demo_data: Include demo records.
        :param services: Start services or only the application containers.
                         This option is incompatible will all the other flags.
        """
        steps = []

        if lock:
            # FIXME: Should this params be accepted? sensible defaults?
            steps.extend(PackagesCommands.lock(pre=True, dev=True))

        if build:
            steps.extend(self.build())

        if services and setup:
            # NOTE: Setup will boot up all service and not bring down
            steps.extend(self.setup(force=True, demo_data=demo_data))
            return steps

        # NOTE: Needed in case there is no setup
        steps.append(
            FunctionStep(
                func=self.docker_helper.start_containers,
                args={"app_only": not services},
                message="Checking if dependencies are locked."
            )
        )

        return steps



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest fixtures."""

from os.path import exists

import pytest
from click.testing import CliRunner

from invenio_cli.cli import cli


@pytest.fixture()
def runner():
    """Click CLI runner."""
    runner = CliRunner()
    with runner.isolated_filesystem():
        yield runner


@pytest.mark.skip()
def test_init(runner):
    """Test init command."""
    result = runner.invoke(cli, ['init'])
    assert result.exit_code == 0
    assert exists('my-site')
    assert exists('my-site/.invenio')


@pytest.mark.skip()
def test_init_with_arg(runner):
    """Test init command."""
    result = runner.invoke(cli, ['init', 'rdm'])
    assert result.exit_code == 0
    assert exists('my-site')
    assert exists('my-site/.invenio')



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

import click


class InvenioCLIConfigError(click.UsageError):
    """Exception when reading/writting from the configuration file."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 Northwestern University.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module commands/commands.py's tests."""

from unittest.mock import patch

import pytest

from invenio_cli.commands import Commands


@pytest.mark.skip()
# NOTE: The patch is the actual import without the `from <package>`
@patch('invenio_cli.helpers.process.run_cmd')
@patch('invenio_cli.commands.commands.DockerHelper')
def test_shell(p_docker_helper, p_run_cmd, mock_cli_config):
    Commands(mock_cli_config).shell()
    p_run_cmd.assert_called_with(['pipenv', 'shell'])


@pytest.mark.skip()
@patch('invenio_cli.helpers.process.run_cmd')
@patch('invenio_cli.commands.commands.DockerHelper')
def test_pyshell(p_docker_helper, p_run_cmd, mock_cli_config):
    Commands(mock_cli_config).pyshell()
    p_run_cmd.assert_called_with(['pipenv', 'run', 'invenio', 'shell'])


@pytest.mark.skip()
@patch('invenio_cli.commands.commands.DockerHelper')
def test_stop(p_docker_helper, mock_cli_config):
    commands = Commands(mock_cli_config, p_docker_helper())
    commands.stop()
    commands.docker_helper.stop_containers.assert_called()


@pytest.mark.skip()
@patch('invenio_cli.helpers.process.run_cmd')
@patch('invenio_cli.commands.commands.DockerHelper')
def test_destroy(p_docker_helper, p_run_cmd, mock_cli_config):
    commands = Commands(mock_cli_config, p_docker_helper())
    commands.destroy()

    p_run_cmd.assert_called_with(['pipenv', '--rm'])
    commands.docker_helper.destroy_containers.assert_called()
    assert mock_cli_config.services_setup is False



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020-2021 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

from os import listdir

from ..helpers.process import ProcessResponse
from .steps import CommandStep


class PackagesCommands(object):
    """Local installation commands."""

    @staticmethod
    def install_packages(packages):
        """Steps to install Python packages.

        It is a class method since it does not require any configuration.
        """
        prefix = ['pipenv', 'run']
        cmd = prefix + ['pip', 'install']
        for package in packages:
            cmd.extend(['-e', package])

        steps = [
            CommandStep(
                cmd=cmd,
                env={'PIPENV_VERBOSITY': "-1"},
                message="Installing python dependencies..."
            )
        ]

        return steps

    @staticmethod
    def outdated_packages():
        """Steps to show outdated packages.

        It is a class method since it does not require any configuration.
        """
        cmd = ['pipenv', 'update', '--outdated']

        steps = [
            CommandStep(
                cmd=cmd,
                env={'PIPENV_VERBOSITY': "-1"},
                message="Checking outdated packages..."
            )
        ]

        return steps

    @staticmethod
    def update_packages():
        """Steps to update all Python packages.

        It is a class method since it does not require any configuration.
        """
        cmd = ['pipenv', 'update']

        steps = [
            CommandStep(
                cmd=cmd,
                env={'PIPENV_VERBOSITY': "-1"},
                message="Updating package(s)..."
            )
        ]

        return steps

    @staticmethod
    def update_package_new_version(package, version):
        """Update invenio-app-rdm version.

        It is a class method since it does not require any configuration.
        """
        prefix = ['pipenv']
        app = prefix + ['install', package + version]

        steps = [
            CommandStep(
                cmd=app,
                env={'PIPENV_VERBOSITY': "-1"},
                message=f"Updating {package} to version {version}..."
            )
        ]

        return steps

    @staticmethod
    def install_locked_dependencies(pre):
        """Installed dependencies from Pipfile.lock using sync."""
        # NOTE: sync has not interactive process feedback
        cmd = ['pipenv', 'sync']
        if pre:
            cmd += ['--pre']

        steps = [
            CommandStep(
                cmd=cmd,
                env={'PIPENV_VERBOSITY': "-1"},
                message="Installing python dependencies... Please be " +
                        "patient, this operation might take some time..."
            )
        ]

        return steps

    @staticmethod
    def lock(pre, dev):
        """Steps lock Python dependencies."""
        cmd = ['pipenv', 'lock']
        if pre:
            cmd += ['--pre']
        if dev:
            cmd += ['--dev']

        steps = [
            CommandStep(
                cmd=cmd,
                env={'PIPENV_VERBOSITY': "-1"},
                message="Locking python dependencies..."
            )
        ]

        return steps

    @staticmethod
    def is_locked():
        """Checks if the dependencies have been locked."""
        locked = 'Pipfile.lock' in listdir('.')
        if not locked:
            return ProcessResponse(
                error="Dependencies were not locked. " +
                      "Please run `invenio-cli packages lock`.",
                status_code=1,
            )

        return ProcessResponse(
            output="Dependencies are locked",
            status_code=0,
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""


import click

from ..commands import Commands, ServicesCommands
from ..helpers.cli_config import CLIConfig
from .utils import pass_cli_config, run_steps


@click.group()
def services():
    """Commands for services management."""


@services.command()
@pass_cli_config
def start(cli_config):
    """Start local services."""
    click.secho("Starting containers...", fg="green")
    commands = ServicesCommands(cli_config)
    steps = commands.start()
    on_fail = "Failed to start services."
    on_success = "Services started successfully."

    run_steps(steps, on_fail, on_success)


@services.command()
@click.option('-f', '--force', default=False, is_flag=True,
              help='Force recreation of db tables, ES indices, queues...')
@click.option('-N', '--no-demo-data', default=False, is_flag=True,
              help='Disable the creation of demo data')
@click.option('--stop-services', default=False, is_flag=True,
              help='Stop containers after setup.')
@click.option('--services/--no-services', '-s/-n',  default=True, is_flag=True,
              help='Enable/disable dockerized services (default: enabled).')
@pass_cli_config
def setup(cli_config, force, no_demo_data, stop_services, services):
    """Setup local services."""
    # no_demo_data = False (default) means "YES to demo_data"
    demo_data = not no_demo_data
    commands = ServicesCommands(cli_config)
    steps = commands.setup(force, demo_data, stop_services, services)
    on_fail = "Failed to setup services."
    on_success = "Successfully setup all services."

    run_steps(steps, on_fail, on_success)


# FIXME: Duplicated code from containers.py
@services.command()
@click.option('-v', '--verbose', default=False, is_flag=True, required=False,
              help='Verbose mode will show all logs in the console.')
@pass_cli_config
def status(cli_config, verbose):
    """Checks if the services are up and running.

    NOTE: currently only ES, DB (postgresql/mysql) and redis are supported.
    """
    commands = ServicesCommands(cli_config)
    services = ["redis", cli_config.get_db_type(), "es"]
    statuses = commands.status(services=services, verbose=verbose)

    messages = [
        {"message": "{}: up and running.", "fg": "green"},
        {"message": "{}: unable to connect or bad response.", "fg": "red"},
        {"message": "{}: no healthcheck function defined.", "fg": "yellow"}
    ]

    for idx, status in enumerate(statuses):
        message = messages[status]
        click.secho(
            message=message.get("message").format(services[idx]),
            fg=message.get("fg")
        )


@services.command()
@pass_cli_config
def stop(cli_config):
    """Stop local services."""
    commands = ServicesCommands(cli_config)
    steps = commands.stop()
    on_fail = "Failed to stop containers."
    on_success = "Stopped containers."

    run_steps(steps, on_fail, on_success)


@services.command()
@pass_cli_config
def destroy(cli_config):
    """Destroy developement services."""
    commands = ServicesCommands(cli_config)
    click.secho("Destroying services' containers, volumes...", fg="green")
    steps = commands.destroy()
    on_fail = "Failed to destroy services' containers."
    on_success = "Services' containers destroyed."

    run_steps(steps, on_fail, on_success)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 CERN.
# Copyright (C) 2019 Northwestern University.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

import os
from pathlib import Path

import click

from ..commands import Commands, ContainersCommands, InstallCommands, \
    LocalCommands, RequirementsCommands, UpgradeCommands
from ..errors import InvenioCLIConfigError
from ..helpers.cli_config import CLIConfig
from ..helpers.cookiecutter_wrapper import CookiecutterWrapper
from .assets import assets
from .containers import containers
from .install import install
from .packages import packages
from .services import services
from .utils import calculate_instance_path, pass_cli_config, run_steps


@click.group()
@click.version_option()
@click.pass_context
def invenio_cli(ctx):
    """Initialize CLI context."""


invenio_cli.add_command(assets)
invenio_cli.add_command(containers)
invenio_cli.add_command(install)
invenio_cli.add_command(packages)
invenio_cli.add_command(services)


@invenio_cli.command('check-requirements')
@click.option('--development', '-d', default=False, is_flag=True,
              help='Check development requirements.')
def check_requirements(development):
    """Checks the system fulfills the pre-requirements."""
    click.secho("Checking pre-requirements...", fg="green")
    steps = RequirementsCommands.check(development)
    on_fail = "Pre requisites not met."
    on_success = "All requisites are fulfilled."

    run_steps(steps, on_fail, on_success)


@invenio_cli.command()
def shell():
    """Shell command."""
    Commands.shell()


@invenio_cli.command()
@click.option('--debug/--no-debug', '-d/', default=False, is_flag=True,
              help='Enable Flask development mode (default: disabled).')
def pyshell(debug):
    """Python shell command."""
    Commands.pyshell(debug=debug)


@invenio_cli.command()
@click.argument('flavour', type=click.Choice(['RDM'], case_sensitive=False),
                default='RDM', required=False)
@click.option('-t', '--template', required=False,
              help='Cookiecutter path or git url to template')
@click.option('-c', '--checkout', required=False,
              help='Branch, tag or commit to checkout if --template is a git url')  # noqa
def init(flavour, template, checkout):
    """Initializes the application according to the chosen flavour."""
    click.secho('Initializing {flavour} application...'.format(
        flavour=flavour), fg='green')

    template_checkout = (template, checkout)
    cookiecutter_wrapper = CookiecutterWrapper(flavour, template_checkout)

    try:
        click.secho("Running cookiecutter...", fg='green')
        project_dir = cookiecutter_wrapper.cookiecutter()

        click.secho("Writing invenio-invenio_cli config file...", fg='green')
        saved_replay = cookiecutter_wrapper.get_replay()
        instance_path = calculate_instance_path()
        CLIConfig.write(project_dir, flavour, saved_replay, instance_path)

        click.secho("Creating logs directory...", fg='green')
        os.mkdir(Path(project_dir) / "logs")

    except Exception as e:
        click.secho(str(e), fg='red')

    finally:
        cookiecutter_wrapper.remove_config()


@invenio_cli.command()
@click.option('--host', '-h',  default='127.0.0.1',
              help='The interface to bind to.')
@click.option('--port', '-p',  default=5000,
              help='The port to bind to.')
@click.option('--debug/--no-debug', '-d/',  default=True, is_flag=True,
              help='Enable/disable debug mode including auto-reloading '
                   '(default: enabled).')
@click.option('--services/--no-services', '-s/-n',  default=True, is_flag=True,
              help='Enable/disable dockerized services (default: enabled).')
@pass_cli_config
def run(cli_config, host, port, debug, services):
    """Starts the local development server.

    NOTE: this only makes sense locally so no --local option
    """
    commands = LocalCommands(cli_config)
    commands.run(host=host, port=str(port), debug=debug, services=services)


@invenio_cli.command()
@pass_cli_config
def destroy(cli_config):
    """Removes all associated resources (containers, images, volumes)."""
    commands = Commands(cli_config)
    services = ContainersCommands(cli_config)
    click.secho(
        "Destroying containers, volumes, virtual environment...", fg="green")
    steps = commands.destroy()  # Destroy virtual environment
    steps.extend(services.destroy())  # Destroy services
    on_fail = "Failed to destroy instance. You can destroy only services " + \
              "using the services command: invenio-cli services destroy"
    on_success = "Instance destroyed."

    run_steps(steps, on_fail, on_success)


@invenio_cli.command()
@click.option('--script', required=True,
              help='The path of custom migration script.'
              )
def upgrade(script):
    """Upgrades the current instance to a newer version."""
    steps = UpgradeCommands.upgrade(script)
    on_fail = "Upgrade failed."
    on_success = "Upgrade sucessfull."

    run_steps(steps, on_fail, on_success)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 Northwestern University.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

import os

from setuptools import find_packages, setup

readme = open('README.rst').read()
history = open('CHANGES.rst').read()

tests_require = [
    'pytest-invenio>=1.4.0',
]

extras_require = {
    'docs': [
        'Sphinx>=3',
    ],
    'tests': tests_require,
}

extras_require['all'] = []
for reqs in extras_require.values():
    extras_require['all'].extend(reqs)


install_requires = [
    'cookiecutter>=1.7.1,<1.8.0',
    'click>=7.1.1,<8.0',
    'click-default-group>=1.2.2,<2.0.0',
    'docker>=4.1.0,<5.0.0',
    'pipenv>=2020.6.2',
    'PyYAML>=5.1.2',
    'pynpm>=0.1.2',
]

packages = find_packages()


# Get the version string. Cannot be done with import!
g = {}
with open(os.path.join('invenio_cli', 'version.py'), 'rt') as fp:
    exec(fp.read(), g)
    version = g['__version__']

setup(
    name='invenio-cli',
    version=version,
    description=__doc__,
    long_description=readme + '\n\n' + history,
    keywords='invenio-cli',
    license='MIT',
    author='CERN & Northwestern University',
    author_email='info@inveniosoftware.org',
    url='https://github.com/inveniosoftware/invenio-cli',
    packages=packages,
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    entry_points={
        'console_scripts': [
            'invenio-cli = invenio_cli.cli.cli:invenio_cli',
        ]
    },
    extras_require=extras_require,
    install_requires=install_requires,
    tests_require=tests_require,
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Development Status :: 3 - Alpha',
    ],
)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

import click
from click_default_group import DefaultGroup

from ..commands import AssetsCommands
from .utils import pass_cli_config, run_steps


@click.group()
def assets():
    """Statics and assets management commands.

    NOTE: The assets commands are intended for development environments.
          To re-build assets on a containers environment use
          `invenio-cli containers build`.
    """


@assets.command()
@click.option('--no-wipe', '-n', default=False, is_flag=True,
              help='Do not remove existing assets.')
@click.option(
    '--production/--development', '-p/-d', default=False, is_flag=True,
    help='Production mode copies files. Development mode symlinks files.'
)
@pass_cli_config
def build(cli_config, no_wipe, production):
    """Build the static and assets files on the local installation."""
    commands = AssetsCommands(cli_config)
    commands.update_statics_and_assets(
        force=not no_wipe,  # If no_wipe=True, it means force=False
        flask_env='production' if production else 'development'
    )


@assets.command()
@click.argument('path', type=click.Path(exists=True))
@pass_cli_config
def install(cli_config, path):
    """Install and link a React module on the local installation."""
    commands = AssetsCommands(cli_config)

    click.secho("Installing React module...", fg="green")
    steps = commands.link_js_module(path)
    on_fail = "Failed to install React module."
    on_success = "React module installed successfully."

    run_steps(steps, on_fail, on_success)


@assets.command('watch')
@pass_cli_config
def watch_assets(cli_config):
    """Watch assets files on the local installation.

    This is the default behaviour when calling `invenio-cli assets watch`.
    """
    commands = AssetsCommands(cli_config)
    commands.watch_assets()


@assets.command('watch-module')
@click.option('--link', '-l', default=False, is_flag=True,
              help='Link the module.')
@click.argument('path', type=click.Path(exists=True))
@pass_cli_config
def watch_module(cli_config, path, link):
    """Watch a React module on the local installation."""
    commands = AssetsCommands(cli_config)
    click.secho("Watching React module...", fg="green")
    steps = commands.watch_js_module(path, link=link)
    on_fail = "Failed set watcher on React module."
    on_success = "Finished watching React module."

    run_steps(steps, on_fail, on_success)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio module to ease the creation and management of applications."""

from .steps import CommandStep


class UpgradeCommands(object):
    """Local installation commands."""

    @staticmethod
    def upgrade(script_path):
        """Steps to perform an upgrade of the invenio instance.

        First, and alembic upgrade is launched to allow alembic to migrate the
        database using SQLAlchemy.
        Then, the custom script is executed.
        Last, the Elasticsearch indexes are destroyed, initialized and rebuilt.
        It is a class method since it does not require any configuration.
        """
        prefix = ['pipenv', 'run', 'invenio']
        alembic_cmd = prefix + ['alembic', 'upgrade']
        destroy_index_cmd = prefix + ['index', 'destroy', '--yes-i-know']
        init_index_cmd = prefix + ['index', 'init']
        rebuild_index_cmd = prefix + ['rdm-records', 'rebuild-index']
        script_cmd = prefix + ['shell', script_path]

        steps = [
            CommandStep(
                cmd=alembic_cmd,
                env={'PIPENV_VERBOSITY': "-1"},
                message="Performing an alembic upgrade..."
            ),
            CommandStep(
                cmd=script_cmd,
                env={'PIPENV_VERBOSITY': "-1"},
                message="Executing data upgrade script..."
            ),
            CommandStep(
                cmd=destroy_index_cmd,
                env={'PIPENV_VERBOSITY': "-1"},
                message="Destroying indexes..."
            ),
            CommandStep(
                cmd=init_index_cmd,
                env={'PIPENV_VERBOSITY': "-1"},
                message="Creating new indexes..."
            ),
            CommandStep(
                cmd=rebuild_index_cmd,
                env={'PIPENV_VERBOSITY': "-1"},
                message="Rebuilding indexes..."
            )
        ]

        return steps



# -*- coding: utf-8 -*-
#
# Copyright (C) 2020 CERN.
# Copyright (C) 2020 Northwestern University.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test environment variables context manager."""

import os

from invenio_cli.helpers.env import env


def test_env():
    """Test environment variables context manager."""
    assert 'FLASK_ENV' not in os.environ
    assert 'FLASK_DEBUG' not in os.environ

    os.environ['FLASK_DEBUG'] = 'true'
    with env(FLASK_ENV='production', FLASK_DEBUG='false'):
        assert os.environ['FLASK_ENV'] == 'production'
        assert os.environ['FLASK_DEBUG'] == 'false'

    assert os.environ['FLASK_DEBUG'] == 'true'
    assert 'FLASK_ENV' not in os.environ



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2020 CERN.
# Copyright (C) 2019-2020 Northwestern University.
#
# Invenio-Cli is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module config_file tests."""
import os
import tempfile
from configparser import ConfigParser
from pathlib import Path
from unittest.mock import patch

import pytest

from invenio_cli.helpers.cli_config import CLIConfig


def test_cli_config_write():
    """Check config file is generated: preliminary superficial test."""

    tmp_dir = tempfile.TemporaryDirectory()
    project_dir = tmp_dir.name
    flavour = 'RDM'
    replay = {
        'cookiecutter': {
            'project_name': 'My Site',
            'project_shortname': 'my-site',
            'project_site': 'my-site.com',
            'github_repo': 'my-site/my-site',
            'description': 'Invenio RDM My Site Instance',
            'author_name': 'CERN',
            'author_email': 'info@my-site.com',
            'year': '2020',
            'database': 'postgresql',
            'elasticsearch': '7',
            '_template': 'https://github.com/inveniosoftware/cookiecutter-invenio-rdm.git'  # noqa
        }
    }
    config_path = Path(project_dir) / CLIConfig.CONFIG_FILENAME
    private_config_path = Path(project_dir) / CLIConfig.PRIVATE_CONFIG_FILENAME

    # No configuration files
    assert not os.path.isfile(config_path)
    assert not os.path.isfile(private_config_path)

    CLIConfig.write(project_dir, flavour, replay)

    assert os.path.isfile(config_path)
    assert os.path.isfile(private_config_path)


@pytest.fixture
def config_dir():
    tmp_dir = tempfile.TemporaryDirectory()
    project_dir = tmp_dir.name
    flavour = 'RDM'
    replay = {
        'cookiecutter': {
            'project_name': 'My Site',
            'project_shortname': 'my-site',
            'project_site': 'my-site.com',
            'github_repo': 'my-site/my-site',
            'description': 'Invenio RDM My Site Instance',
            'author_name': 'CERN',
            'author_email': 'info@my-site.com',
            'year': '2020',
            'database': 'postgresql',
            'elasticsearch': '7',
            '_template': 'https://github.com/inveniosoftware/cookiecutter-invenio-rdm.git'  # noqa
        }
    }
    # Need to yield in order for tmp_dir to not be gc'ed and therefore the
    # temporary directory deleted
    yield CLIConfig.write(project_dir, flavour, replay)


@pytest.mark.skip()
@patch("invenio_cli.helpers.cli_config.exit")
def test_cli_config_config_files_not_found(patched_exit, config_dir):
    wrong_path = tempfile.TemporaryDirectory()

    cli_config = CLIConfig(wrong_path.name)

    patched_exit.assert_called_with(1)


def test_cli_config_get_project_dir(config_dir):
    cli_config = CLIConfig(config_dir)

    assert cli_config.get_project_dir() == config_dir


def test_cli_config_instance_path(config_dir):
    cli_config = CLIConfig(config_dir)

    assert cli_config.get_instance_path() == Path('')

    # Update instance path to now see if we retrieve it
    instance_path = os.path.join(config_dir, '.venv/')
    cli_config.update_instance_path(instance_path)

    assert cli_config.get_instance_path() == Path(instance_path)


def test_cli_config_services_setup(config_dir):
    cli_config = CLIConfig(config_dir)

    assert cli_config.get_services_setup() is False

    cli_config.update_services_setup(True)

    assert cli_config.get_services_setup() is True


def test_cli_config_get_project_shortname(config_dir):
    cli_config = CLIConfig(config_dir)

    assert cli_config.get_project_shortname() == 'my-site'
